#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 13 07:50:55 2019

@author: fuaa
"""
from matplotlib import pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.cm as cm
import numpy as np
import getpass
import socket

_THRESHOLD = 10e-9

#########################################################################
def identificar_parcial():
    username = getpass.getuser()
    hostname = socket.gethostname()
    print("Usuario %s en %s." % (username, hostname))


##################################################################################
def printcolor( mensaje, color="k" ):
    """
    Función para imprimir en colores y con el formato de interés.
    """
    if   (color == "r"): mensajeColor = "\x1b[31m" + mensaje + "\x1b[0m"
    elif (color == "g"): mensajeColor = "\x1b[32m" + mensaje + "\x1b[0m"
    elif (color == "y"): mensajeColor = "\x1b[33m" + mensaje + "\x1b[0m"
    elif (color == "b"): mensajeColor = "\x1b[34m" + mensaje + "\x1b[0m"
    elif (color == "p"): mensajeColor = "\x1b[35m" + mensaje + "\x1b[0m"
    elif (color == "c"): mensajeColor = "\x1b[36m" + mensaje + "\x1b[0m"
    else: mensajeColor = mensaje
    mensaje_out = " " + mensajeColor 
    print ( mensaje_out )


##################################################################################
def fuaa_assert(condicion, 
                mensajeFalse = "El resultado no es válido.", 
                mensajeTrue = "Resultado validado." ):
    """
    Función similar al assert.
    """
    # Custom assert.
    if ( condicion ):
        printcolor( mensajeTrue, "g" )
    else:
        printcolor( mensajeFalse, "r" )


##################################################################################
def son_iguales(x1, x2, threshold = _THRESHOLD):
    """
    Evaluar si dos elementos son iguales o no, con una tolerancia dada (threshold).
    """
    if x1.shape == x2.shape:
        if isinstance(x1, np.ndarray):
            dif = np.sqrt(np.sum( ( x1 - x2 )**2 )) / x1.size
        elif isinstance(x1, float):
            dif = np.abs( x1 - x2 )
        condicion = (dif < threshold)
    else:
        condicion = False
    return condicion



##################################################################################
def visualizar_conjunto(X, y, titulo='Conjunto de datos' ,dpi=300):

    '''
    Entrada: 
        X: matriz de tamaño Nx3 que contiene el conjunto de puntos de entrenamiento expresados en
           en coordenadas homogeneas. La primera coordenada de cada punto es uno.
        y: etiquetas asignadas a los puntos
    '''
    plt.figure(dpi)

    # Se grafican los puntos sorteados
    plt.scatter(X[y == -1, 1], X[y == -1, 2], s=40,
                color='r', marker='*', label='etiqueta -1')
    plt.scatter(X[y == 1, 1], X[y == 1, 2], s=40,
                color='b', marker='o', label='etiqueta 1')

    plt.legend()
    plt.axis('equal')
    plt.xlabel('x1')
    plt.ylabel('x2')
    plt.title(titulo)


#########################################################################
def visualizar_modelo(X, y, w_g, fig_size=(7, 7), title='Modelo lineal encontrado'):
    '''
    Entrada: 
        X: matriz de tamaño Nx3 que contiene el conjunto de puntos de entrenamiento expresados en
           en coordenadas homogeneas. La primera coordenada de cada punto es uno.
        y: etiquetas asignadas a los puntos
        w_g: parámetros del modelo lineal encontrados 
    '''
    plt.figure(figsize=fig_size)

    # Se grafican los puntos sorteados
    plt.scatter(X[y == -1, 1], X[y == -1, 2], s=40,
                color='r', marker='*', label='etiqueta -1')
    plt.scatter(X[y == 1, 1], X[y == 1, 2], s=40,
                color='b', marker='o', label='etiqueta 1')

    x1_min = X[:, 1].min()
    x1_max = X[:, 1].max()
    x1 = np.linspace(x1_min, x1_max)

    # Plot models
    if w_g[2] == 0:
        x2_g = -(w_g[0]/w_g[2])*np.ones(x1.shape)
    else:
        # Se grafica la superficie de decisión encontrada
        x2_g = - w_g[1]/w_g[2] * x1 + -w_g[0]/w_g[2]
    plt.plot(x1, x2_g, label='Funcion encontrada con modelo', color='red')

    y1_min = X[:, 2].min()
    y1_max = X[:, 2].max()

    plt.legend()
    plt.axis('equal')
    plt.xlabel('x1')
    plt.ylabel('x2')
    plt.title(title)
    plt.xlim(x1_min, x1_max)
    plt.ylim(y1_min, y1_max)


##################################################################################

def visualizar_multiples_modelos (X, y, w_list, alphas ,fig_size=(7,7), dpi=300):
    
    '''
    Entrada: 
        X: matriz de tamaño Nx3 que contiene el conjunto de puntos de entrenamiento expresados en
           en coordenadas homogeneas. La primera coordenada de cada punto es uno.
        y: etiquetas asignadas a los puntos
        w_g: parámetros del modelo lineal encontrados 
    '''
    plt.figure(figsize=fig_size, dpi=dpi)

    # Se grafican los puntos sorteados
    plt.scatter(X[y == -1, 1], X[y == -1, 2], s=40,
                color='r', marker='*', label='etiqueta -1')
    plt.scatter(X[y == 1, 1], X[y == 1, 2], s=40,
                color='b', marker='o', label='etiqueta 1')

    x1_min = X[:, 1].min()
    x1_max = X[:, 1].max()
    x1 = np.linspace(x1_min, x1_max)

    normColor = mcolors.Normalize(vmin=alphas.min(), vmax=alphas.max())
    colormap = cm.jet

    for i, w_g in enumerate(w_list):
        # Plot models
        if w_g[2] == 0:
            x2_g = -(w_g[0]/w_g[2])*np.ones(x1.shape)
        else:
            # Se grafica la superficie de decisión encontrada
            x2_g = - w_g[1]/w_g[2] * x1 + -w_g[0]/w_g[2]
        plt.plot(x1, x2_g, label='Modelo lineal encontrado (alpha = %.02f)' % alphas[i],
                 color=colormap(normColor(alphas[i])))

    y1_min = X[:, 2].min()
    y1_max = X[:, 2].max()

    plt.legend(bbox_to_anchor=(1.02, 1.05))
    plt.axis('equal')
    plt.xlabel('x1')
    plt.ylabel('x2')
    plt.xlim(x1_min, x1_max)
    plt.ylim(y1_min, y1_max)


#########################################################################
def comparar_modelos(X, y, ws, labels=None):
    '''
    Entrada: 
        X: matriz de tamaño Nx3 que contiene el conjunto de puntos de entrenamiento expresados en
           en coordenadas homogeneas. La primera coordenada de cada punto es uno.
        y: etiquetas asignadas a los puntos
        w_g: parámetros del modelo lineal encontrados 
    '''
    plt.figure(figsize=(7, 7))

    # Se grafican los puntos sorteados
    plt.scatter(X[y == -1, 1], X[y == -1, 2], s=40,
                color='r', marker='*', label='etiqueta -1')
    plt.scatter(X[y == 1, 1], X[y == 1, 2], s=40,
                color='b', marker='o', label='etiqueta 1')

    x1_min = X[:, 1].min()
    x1_max = X[:, 1].max()
    x1 = np.linspace(x1_min, x1_max)

    if labels is None:
        labels = ['Modelo %d' % i for i in range(len(ws))]

    for i, w in enumerate(ws):
        # Plot models
        if w[2] == 0:
            x2_g_1 = -(w[0]/w[2])*np.ones(x1.shape)
        else:
            # Se grafica la superficie de decisión encontrada
            x2_g_1 = - w[1]/w[2] * x1 + -w[0]/w[2]
        plt.plot(x1, x2_g_1, label=labels[i])

    # if w_g_2[2]==0:
    #    x2_g_2 = -(w_g_2[0]/w_g_2[2])*np.ones(x1.shape)
    # else:
    #    # Se grafica la superficie de decisión encontrada
    #    x2_g_2 = - w_g_2[1]/w_g_2[2] * x1 + -w_g_2[0]/w_g_2[2]
    #plt.plot(x1, x2_g_2, label = 'Funcion encontrada con Modelo 2', color='blue')

    y1_min = X[:, 2].min()
    y1_max = X[:, 2].max()

    plt.legend()
    plt.axis('equal')
    plt.xlabel('x1')
    plt.ylabel('x2')
    plt.title('Modelo lineal encontrado')
    plt.xlim(x1_min, x1_max)
    plt.ylim(y1_min, y1_max)

##################################################################################
def validar_resultado(*args, **kwargs):
    """"
    Función para validar resultado a invocar desde el notebook.
    """
    # _ES_CORRECTO = False
    _DEBUG = False

    # Abrir el cartel del validación.
    print( "+-------------------------------------------------------------------------+" )
    print( "|               FuAA (1er. parcial 2022): validar resultado               |" )
    print( "+-------------------------------------------------------------------------+" )
    for key, value in kwargs.items():
        if key == "debug":
            _DEBUG = value
    if _DEBUG: 
        print('args:', args)
        print('kwargs:', kwargs)

    if ( len(args) == 0 ):
        print( "| Sin opciones para evaluar.                                              |" )
        print( "+-------------------------------------------------------------------------+" )
        return False
    
    ###########################################################
    # Ejercicio 2. Costo regresión logística
    ###########################################################
    elif (args[0] == "calcular_costo"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return

        np.random.seed(33)
        X5 = np.random.randn(3,4)
        X5[:,0]=1
        y5 = np.array([1,-1,1])
        w5 = np.random.randn(4)

        c5 = test_algoritmo(X5, y5, w5 )
        # print(c5)
        condicion1 =  np.abs(c5 - 0.13048058961396589) < _THRESHOLD
        fuaa_assert(   condicion1, \
                    mensajeFalse = "Revisar la implementación del costo de la regresión logística", \
                    mensajeTrue = "Implementación del costo de la regresión logística: resultado validado.")

    ###########################################################
    # Ejercicio 2. Gradiente regresión logística
    ###########################################################
    elif (args[0] == "calcular_gradiente"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return

        np.random.seed(33)
        X5 = np.random.randn(3,4)
        X5[:,0]=1
        y5 = np.array([1,-1,1])
        w5 = np.random.randn(4)

        g5 = test_algoritmo(X5,y5,w5) 
        condicion1  = g5.shape == w5.shape
        fuaa_assert(condicion1, \
                    mensajeFalse = "Revisar las dimensiones del gradiente", \
                    mensajeTrue = "Dimensiones del gradiente son correctas.")
        condicion2 = son_iguales(g5, np.array([-0.10645973,  0.10632848,  0.12291721,  0.07835299]))
        fuaa_assert(   condicion2, \
                    mensajeFalse = "Revisar la implementación del gradiente de la regresión logística", \
                    mensajeTrue = "Implementación del gradiente de la regresión logística: resultado validado.")

    ###########################################################
    # Ejercicio 2. Costo regresión logística ponderada
    ###########################################################
    elif (args[0] == "calcular_costo_ponderado"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return

        np.random.seed(42)
        X = np.random.randn(3,4)
        X[:,0]=1
        y = np.array([1,-1,1])
        w = np.random.randn(4)
        weights = np.empty(len(y))
        weights[y == 1] = 0.2
        weights[y == -1] = 0.8

        cost = test_algoritmo(X, y, w, weights )
        # print(cost)
        condicion1 =  np.abs(cost - 0.1713398927737643) < _THRESHOLD
        fuaa_assert(   condicion1, \
                    mensajeFalse = "Revisar la implementación del costo de la regresión logística ponderada", \
                    mensajeTrue = "Implementación del costo de la regresión logística ponderada: resultado validado.")

    ###########################################################
    # Ejercicio 2. Gradiente regresión logística ponderada
    ###########################################################
    elif (args[0] == "calcular_gradiente_ponderado"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return

        np.random.seed(42)
        X = np.random.randn(3,4)
        X[:,0]=1
        y = np.array([1,-1,1])
        w = np.random.randn(4)
        weights = np.empty(len(y))
        weights[y == 1] = 0.2
        weights[y == -1] = 0.8

        g = test_algoritmo(X,y,w, weights) 
        # print(g)
        condicion1  = g.shape == w.shape
        fuaa_assert(condicion1, \
                    mensajeFalse = "Revisar las dimensiones del gradiente", \
                    mensajeTrue = "Dimensiones del gradiente son correctas.")
        condicion2 = son_iguales(g, np.array([-0.0622332, -0.01310058,  0.01129954, -0.05300562]))
        fuaa_assert(   condicion2, \
                    mensajeFalse = "Revisar la implementación del gradiente de la regresión logística ponderada", \
                    mensajeTrue = "Implementación del gradiente de la regresión logística ponderada: resultado validado.")
    ###########################################################
    # Ejercicio 2. Regresión Lineal
    ###########################################################
    elif (args[0] == "regresion_lineal"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return
        
        np.random.seed(43)
        y3 = np.random.randn(9)
        Z3 = np.random.randn(9,7)
        Z3[:,0]=1
        w3 = test_algoritmo(Z3, y3 )
        w3_true = np.array([ 0.99480683, -1.02402373,  0.40146952,  0.21686739,  0.23815722, -1.23370181,
 -0.64320162])
        condicion1 = son_iguales( w3, w3_true )
        fuaa_assert(   condicion1, \
                    mensajeFalse = "Revisar la implementación de la regresión lineal", \
                    mensajeTrue = "Implementación de la regresión lineal: resultado validado.")

    ###########################################################
    # Ejercicio 2. Regresión Ridge
    ###########################################################
    elif (args[0] == "regresion_Ridge"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return

        np.random.seed(43)
        y3 = np.random.randn(9)
        Z3 = np.random.randn(9,7)
        Z3[:,0] = 1
        reg = 3.3
        w3 = test_algoritmo(Z3, y3, reg )
        w3_true = np.array([ 0.58610027, -0.09425186,  0.12976429,  0.11544501,  0.14085992, -0.39307442,
 -0.20333349])
        condicion1 = son_iguales( w3, w3_true )
        fuaa_assert(   condicion1, \
                    mensajeFalse = "Revisar la implementación de la regresión lineal regularizada", \
                    mensajeTrue= "Implementación de la regresión lineal regularizada: resultado validado.")

    ###########################################################
    # Ejercicio 2. Obtener hipótesis.
    ###########################################################
    elif (args[0] == "obtener_hipotesis"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return

        np.random.seed(42)
        K = 10
        x1 = np.random.randn(K)
        y1 = np.random.randn(K)
        x2 = np.random.randn(K)
        y2 = np.random.randn(K)
        a, b = test_algoritmo(x1,y1,x2,y2,'H0')
        condicion_a = son_iguales( a, np.zeros(K) )
        condicion_b = son_iguales( b, np.array([-0.53256215, 0.69327422, 0.11423252, -1.48549559, -0.45118646, -0.89156559, -0.40198376, -0.8227114, -1.11810506, -0.60772123]) )
        fuaa_assert( condicion_a, 
                    mensajeFalse = "Parámetros 'a' con H0 no validados.", 
                    mensajeTrue = "Parámetros 'a' con H0 validados." )
        fuaa_assert( condicion_b,
                    mensajeFalse = "Parámetros 'b' con H0 no validados.", 
                    mensajeTrue = "Parámetros 'b' con H0 validados." )

        a, b = test_algoritmo(x1,y1,x2,y2,'H1')
        condicion_a = son_iguales( a, np.array([-0.14272265, -26.48788688, 0.44032569, -0.29024211, -8.21154654, -1.90852891, -0.44747339, 5.80470861, 3.2033256, -1.92886739]) )
        condicion_b = son_iguales( b, np.array([-0.39252533, -4.12805892, -0.04323163, -1.47123285, -3.64767917, -1.00914468, -0.30617541, -4.14048764, 0.59585525, -0.36577733]) )
        fuaa_assert( condicion_a,
                    mensajeFalse = "Parámetros 'a' con H1 no validados.", 
                    mensajeTrue = "Parámetros 'a' con H1 validados." )
        fuaa_assert( condicion_b,
                    mensajeFalse = "Parámetros 'b' con H1 no validados.", 
                    mensajeTrue = "Parámetros 'b' con H1 validados." )

    ###########################################################
    # Ejercicio 2. MSE.
    ###########################################################
    elif (args[0] == "mse"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return

        np.random.seed(43)
        N = 10
        d = 4
        X = np.random.rand(N,d+1)
        w = np.random.rand(d+1)
        y = np.random.rand(N)
        err = test_algoritmo(X,w,y)
        condicion = np.abs(err - 1.2015737362409753) < _THRESHOLD #(err == 1.2015737362409753)
        fuaa_assert( condicion,
                    mensajeFalse = "El MSE calculado no es correcto.",
                    mensajeTrue = "Implementación de mse(): resultado validado." )


    ###########################################################
    # Ejercicio 2. Calculo de sesgo y varianza
    ###########################################################
    elif (args[0] == "calcular_sesgo_varianza"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return

        np.random.seed(43)
        Ntest = 10
        K = 14
        X = np.random.rand( Ntest )
        a = np.random.rand( K )
        b = np.random.rand( K )
        sesgo, varianza = test_algoritmo( X, a, b )

        condicion_sesgo = np.abs(sesgo - 0.07425087677513127) <_THRESHOLD #( sesgo == 0.07425087677513127 )
        fuaa_assert( condicion_sesgo,
                    mensajeFalse = "El 'sesgo' calculado no es correcto.",
                    mensajeTrue = "Implementación de cálculo del sesgo: resultado validado." )

        condicion_varianza = np.abs(varianza - 0.14012245581195754)  <_THRESHOLD #( varianza == 0.14012245581195754 )
        fuaa_assert( condicion_varianza,
                    mensajeFalse = "La 'varianza' calculado no es correcto.",
                    mensajeTrue = "Implementación de cálculo de la varianza: resultado validado." )

    ###########################################################
    # No hay ninguna opción de ejercicio.
    ###########################################################
    else:
        printcolor( "Ninguna opción revisada." ) 

    # Cerrar el cartel.
    print( "+-------------------------------------------------------------------------+" )
            
# condicion = False
# mensaje = "Este ese el texto a mostrar en caso de condición falsa."
# validar_resultado( "test", condicion, mensaje )


